# This code is from a tutorial put together by Dan Aloni
# http://blog.aloni.org/posts/backprop-with-tensorflow
# modified to use dataset.py by Ben Sattelberg ... April 2019

import tensorflow as tf

# Use the dataset from https://github.com/tensorflow/models/tree/master/official/mnist/dataset.py
import dataset

# Construct the dataset and an iterator
train = dataset.train(directory='MNIST_data')
train = train.cache() # Store in memory
# This shuffles the dataset and tells it to repeat sampling
train = train.apply(tf.data.experimental.shuffle_and_repeat(buffer_size=50000, count=-1))
# Batch size of 10
train = train.batch(10)
# Create an iterator that goes through the dataset
iterTrain = train.make_initializable_iterator()
# This evaluates to the next object on each sess.run
el = iterTrain.get_next()

# Construct the test set and an iterator
test = dataset.test(directory='MNIST_data')
test = test.cache()
test = test.apply(tf.data.experimental.shuffle_and_repeat(buffer_size=50000, count=-1))
test = test.batch(1000)
iterTest = test.make_initializable_iterator()
elTest = iterTest.get_next()

# Since the input and output vectors are tensors now, we need to verify 
# if we are training or not
isTrain = tf.placeholder(tf.bool, shape=())

# Can't use if statements
# uses the training iterator when isTrain is true
# and the testing iterator otherwise
a_0, y = tf.cond(isTrain, lambda: el, lambda: elTest)
# Need to do one hot encoding on y
y = tf.one_hot(y, 10)
# Convert y to a float for later calculations
y = tf.dtypes.cast(y, tf.float32)

middle = 30
w_1 = tf.Variable(tf.truncated_normal([784, middle]))
b_1 = tf.Variable(tf.truncated_normal([1, middle]))
w_2 = tf.Variable(tf.truncated_normal([middle, 10]))
b_2 = tf.Variable(tf.truncated_normal([1, 10]))

def sigma(x):
    return tf.math.divide(tf.constant(1.0),
                  tf.add(tf.constant(1.0), tf.exp(tf.negative(x))))

z_1 = tf.add(tf.matmul(a_0, w_1), b_1)
a_1 = sigma(z_1)
z_2 = tf.add(tf.matmul(a_1, w_2), b_2)
a_2 = sigma(z_2)

diff = tf.subtract(a_2, y)

def sigmaprime(x):
    return tf.multiply(sigma(x), tf.subtract(tf.constant(1.0), sigma(x)))

d_z_2 = tf.multiply(diff, sigmaprime(z_2))
d_b_2 = d_z_2
d_w_2 = tf.matmul(tf.transpose(a_1), d_z_2)

d_a_1 = tf.matmul(d_z_2, tf.transpose(w_2))
d_z_1 = tf.multiply(d_a_1, sigmaprime(z_1))
d_b_1 = d_z_1
d_w_1 = tf.matmul(tf.transpose(a_0), d_z_1)


eta = tf.constant(0.5)
step = [
    tf.assign(w_1, tf.subtract(w_1, tf.multiply(eta, d_w_1)))
  , tf.assign(b_1, tf.subtract(b_1, tf.multiply(eta,
                               tf.reduce_mean(d_b_1, axis=[0]))))
  , tf.assign(w_2, tf.subtract(w_2, tf.multiply(eta, d_w_2)))
  , tf.assign(b_2, tf.subtract(b_2, tf.multiply(eta,
                               tf.reduce_mean(d_b_2, axis=[0]))))
]

acct_mat = tf.equal(tf.argmax(a_2, 1), tf.argmax(y, 1))
acct_res = tf.reduce_sum(tf.cast(acct_mat, tf.float32))

sess = tf.InteractiveSession()
sess.run(tf.global_variables_initializer())
sess.run(iterTrain.initializer)
sess.run(iterTest.initializer)

for i in range(10000):
    sess.run(step, feed_dict = {isTrain: True})
    if i % 1000 == 0:
        res = sess.run(acct_res, feed_dict = {isTrain: False})
        print(res)
